/* 
 * CSE142, Summer 1999
 * Homework 5
 * Whack-A-Mole
 *
 *
 */

#include "gp142.h"
#include <stdio.h>
#include <math.h>

#include <time.h>
	/* 
	* <time.h> is included so the current time can be accessed
	* to initialize the random number generator.  If you don't
	* use random numbers (and you don't have to) you can either
	* ignore this #include or delete it.
	*/

/* Logical constants */
#define TRUE    		1
#define FALSE   		0

/* 
 * For many of the GP142 functions, a line width of
 * zero means to fill the shape being drawn.  The
 * constant FILL is used in function calls when a solid
 * shape is desired.
 */
#define FILL 0



/*************** Menu buttons: ************************/
/* Top-left coordinates of button palette */
#define MENU_TOP		(GP142_YMAX - 5)
#define MENU_LEFT		(-GP142_XMAX + 5)

/* Height and width of buttons in pixels */
#define BUTTON_HEIGHT  	25
#define BUTTON_WIDTH 	110

/* Number of buttons */
#define NUM_BUTTONS 	2


/*********** Possible locations of mouse clicks: ******/
/*   Mouse click in some button */
#define BUTTON_CLICK 1

/*   Mouse click elsewhere */
#define OTHER_CLICK 0


/* Board specifications */
#define NUM_ROWS    4
#define NUM_COLUMNS 4

#define GAP 2

#define CELL_WIDTH (GP142_XMAX / 4)
#define CELL_HEIGHT (GP142_XMAX / 4)

#define BOARD_WIDTH (CELL_WIDTH * NUM_COLUMNS + GAP * (NUM_COLUMNS-1))
#define BOARD_HEIGHT (CELL_HEIGHT * NUM_ROWS + GAP * (NUM_ROWS-1))

#define BOARD_MINX   (-BOARD_WIDTH /2 )
#define BOARD_MINY   (-BOARD_HEIGHT / 2)

#define BOARD_MAXX   (BOARD_MINX + BOARD_WIDTH)
#define BOARD_MAXY	 (BOARD_MINY + BOARD_HEIGHT)


#define MOLE_SIZE 25


/********* Structures ***************/
typedef struct {
	int bad_mole;     /* Bad moles make you lose a life */
	int duration;     /* How long the mole will last on the screen */
	int mole_present; /* If there's a mole here */
	int been_hit;     /* If a mole has been hit, draw a hit icon */
} mole; 


/*****************************************************************************

     FUNCTION PROTOTYPES

*****************************************************************************/

/* Clears the screen so we can redraw it */
void clear_screen();

/* Draw the buttons */
void draw_buttons();

/* Draws the playing field */
void draw_board();



/* Print our score and number of lives*/
void draw_status_message(int score, int lives);

/* Return kind of mouse click (in a button, etc.) */
int  classify_mouse_click(int mouse_x, int mouse_y);

/* Take care of the user clicking over one of the buttons */
void handle_button(int x, int y, mole mole_board[NUM_ROWS][NUM_COLUMNS], 
				   int *lives, int *score, int *quit);

/* Process mouse click somewhere on the canvas */
void handle_canvas(int x, int y, mole moles[NUM_ROWS][NUM_COLUMNS],
				   int *lives, int *score);

/* Display error message if internal error is detected. */
void something_is_wrong (void);

/* Initialize the array to have no moles. */
void initialize_board(mole mole_board[NUM_ROWS][NUM_COLUMNS]);

/* Add a mole to a random cell in the board */
void add_mole(mole mole_board[NUM_ROWS][NUM_COLUMNS]);

/* Decrement the timers for all of the moles. If a moles timer
   runs out, erase him. */
void handle_mole_timer(mole mole_board[NUM_ROWS][NUM_COLUMNS]);

/* Draw all of the moles on the board */
void draw_moles(mole mole_board[NUM_ROWS][NUM_COLUMNS]);

/* Draw a mole centered about the point (x,y). */
void draw_mole(int x, int y, int bad_mole);

/* Draw a whacked mole */
void draw_hit_mole(int x, int y);




/*****************************************************************************

     MAIN

*****************************************************************************/

/*
 * Main program contents:
 *   - Major program variables that persist throughout execution
 *   - Event loop that processes GP142 events until the user quits,
 *     and updates variables and redraws the window as needed.
 */
int main(void)
{
	/* Event handling and buttons: */
	int  quit;				/* = "user has selected quit"				*/
	int  event;				/* the most recent GP142 event				*/
	char key_pressed;		/* last keyboard character from key event	*/
	int  mouse_x, mouse_y;	/* x and y coordinates of latest mouse event*/
	int  click_kind;		/* kind of mouse click (button, etc.)		*/



	/* User defined variables */
	mole mole_board[NUM_ROWS][NUM_COLUMNS];
	int timer = 10;
	int lives = 3;
	int score = 0;


	/* Initialize graphics package, clear the window,	*/
	/* hide the text window, turn logging off, and start animation.	*/
	GP142_open();
	GP142_show_text(0);     
	GP142_logging(LOG_OFF);
	GP142_clear();
	GP142_animate(ANI_RUN);

	/* Seed the random number generator (to make sure we get
	 * some good, random numbers)
	 */
	srand(time(NULL));

	/* Set each element of the array to having no mole. */
	initialize_board(mole_board);

	mole_board[0][0].mole_present = TRUE;

	/*
	 * Main event loop:
	 * ---- ----- -----
	 * Wait for the next user action, decode it, and call an
	 * appropriate function to handle it.  Repeat until the
	 * user selects quit.
	 */
	quit = FALSE;
	while (!quit) {
		
		/* get next event */
		event = GP142_await_event(&mouse_x, &mouse_y, &key_pressed);
		
		/* perform appropriate action depending on kind of event */
		switch (event) 
		{
			case GP142_QUIT:
				/* Quit selected */
				quit = TRUE;
				break;
				
			case GP142_KBD:
				/* Key pressed.  Not used now.  If the program should		*/
				/* react to keyboard events, insert something appropriate	*/
				/* here.  Note that scanf CANNOT be used with GP142.		*/
				break;
				
			case GP142_MOUSE:
				/* Mouse click.  Determine if in a menu button	*/
				/* or elsewhere and process accordingly.		*/
				click_kind = classify_mouse_click(mouse_x, mouse_y);

				if (click_kind == BUTTON_CLICK) 
				{
					handle_button(mouse_x, mouse_y, mole_board, &lives, &score, &quit);
				} 
				else 
				{
					handle_canvas(mouse_x, mouse_y, mole_board, &lives, &score);
				}
				break;
				
			case GP142_PERIODIC:
				if (lives > 0)
				{
					/* Tick down the timer */
					timer = timer - 1;
					
					/* Decreases the timers for each mole */
					handle_mole_timer(mole_board);
					
					if (timer <= 0)
					{
						/* Increment the timer by up to 20 ticks (about 2 seconds) */
						timer = rand()%20;
						
						/* Now put a new mole on the board */
						add_mole(mole_board);
					}
				}
				break;
				
			default:
				/* unknown event */
				break;
				
		}  /*end switch */
		
		/******* Redraw everything ******/

		/* Draw a white rectangle over the entire screen. Otherwise
		 * things won't get erased properly */
		clear_screen();
		draw_buttons();

		draw_board();
		draw_moles(mole_board);
		draw_status_message(score, lives);

		if (lives <= 0)
		{
			GP142_printfXY(RED, BOARD_MINX, 0, 50, "GAME OVER!!!");
		}
		
	}  /* end event loop */
	
	
	/* Termination: shut down graphics window and exit */
	GP142_close();
	return 0;
}  /*end main */


/*****************************************************************************

    DRAWING ROUTINES

*****************************************************************************/

/* Clears the screen by drawing a white rectangle over it. If we don't do this,
 * images won't properly be erased. 
 */
void clear_screen()
{
	GP142_rectangleXY(WHITE, -GP142_XMAX, -GP142_YMAX,
		                          GP142_XMAX, GP142_YMAX,
								  FILL);
}


/*
 * Draw brown squares for the board.
 */
void draw_board()
{
	int row, col;
	int x1, y1;
	/* Draw a bunch of squares in a grid */
	for (row = 0; row < NUM_ROWS; row++)
	{
		for (col = 0; col < NUM_COLUMNS; col++)
		{
			x1 = BOARD_MINX + col*CELL_WIDTH;
			y1 = BOARD_MINY + row*CELL_HEIGHT;
			GP142_rectangleXY(BROWN, x1 + GAP, y1 + GAP, 
								 x1 + CELL_WIDTH - GAP, y1 + CELL_HEIGHT - GAP,
								 FILL);
		}
	}
	/* Draw a black border around the board */
	GP142_rectangleXY(BLACK, BOARD_MINX - 3, BOARD_MINY - 3,
							 BOARD_MAXX - 3, BOARD_MAXY - 3, 6);
} 


/* Draw all of the moles. Loop through the array and draw
 * any mole that you find.
 */
void draw_moles(mole mole_board[NUM_ROWS][NUM_COLUMNS])
{
	int i,j;
	int x_coord, y_coord;

	for (i = 0; i < NUM_ROWS; i++)
		for (j = 0; j < NUM_COLUMNS; j++)
		{
			if (mole_board[i][j].mole_present == TRUE)
			{
				x_coord = BOARD_MINX + (i * CELL_WIDTH) + CELL_WIDTH/2;
				y_coord = BOARD_MINY + j * CELL_HEIGHT  + CELL_HEIGHT/2;
				if (mole_board[i][j].been_hit == FALSE)
					draw_mole(y_coord, x_coord, mole_board[i][j].bad_mole);
				else
					draw_hit_mole(y_coord, x_coord);
			}
		}
}



/* Draw a mole centered about the point (x,y). */
void draw_mole(int x, int y, int bad_mole)
{
	int color;

	if (bad_mole) color = RED;
	else color = DUSTY_PLUM;
	/* Ears */
	GP142_ovalXY(color, x - 13, y + 12, x - 30, y + 35, 0);
	GP142_ovalXY(color, x + 13, y + 12, x + 30, y + 35, 0);

	GP142_ovalXY(BLACK, x - 15, y + 14, x - 28, y + 33, 1);
	GP142_ovalXY(BLACK, x + 15, y + 14, x + 28, y + 33, 1);

	/* Draw the head */
	GP142_circleXY(color, x, y, MOLE_SIZE);

	/* Draw the eyes */
	GP142_ovalXY(WHITE, x + 7, y + 4, x + 12, y + 12, 5);
	GP142_ovalXY(WHITE, x - 7, y + 4, x - 12, y + 12, 5);

	/* Eyeballs */
	GP142_circleXY(BLACK, x + 9, y + 8, 3);
	GP142_circleXY(BLACK, x - 9, y + 8, 3);

	/* Nose */
	GP142_ovalXY(BLACK, x - 7, y - 9, x + 7, y - 1, 0);

	/* Mouth */
	GP142_ovalXY(PURPLE, x - 15, y - 12, x + 15, y - 18, 4);
}


/*
 * Draw a yellow circle with the word BAM in it 
 */
void draw_hit_mole(int x, int y)
{
	GP142_circleXY(YELLOW, x, y, 22);
	GP142_printfXY(RED, x - 15, y-7, 15, "BAM!");
}



/*
 * Prints out the name of the game (Whack a Mole), number of lives
 * remaining, and your score.
 */
void draw_status_message(int score, int lives)
{

	int score_x = -GP142_XMAX + BUTTON_WIDTH +60;
	int score_y = -GP142_YMAX + 45;
	int lives_x = -GP142_XMAX + BUTTON_WIDTH + 60;
	int lives_y = -GP142_YMAX + 20;

	int font_size = 22;

	/* Draw a colorful display! */
	GP142_printfXY(GREEN, BOARD_MINX-10, BOARD_MAXY + 20, 50, "Whack");
	GP142_printfXY(ORANGE, BOARD_MINX + 140, BOARD_MAXY + 20, 50, "  A  ");
	GP142_printfXY(ICE_BLUE, BOARD_MINX + 190, BOARD_MAXY + 20, 50, "  Mole");
	GP142_printfXY(RED, BOARD_MINX + 325, BOARD_MAXY + 20, 50, "!");
	GP142_printfXY(BLUE, BOARD_MINX + 340, BOARD_MAXY + 20, 50, "!");


	GP142_printfXY(BLUE, score_x, score_y, font_size, "Score: %d", score);
	GP142_printfXY(RED, lives_x, lives_y, font_size, "Lives: %d", lives);
	
}


/* Initialize the array to have no moles. */
void initialize_board(mole mole_board[NUM_ROWS][NUM_COLUMNS])
{
	int i, j;

	for (i = 0; i < NUM_ROWS; i++)
		for (j = 0; j < NUM_COLUMNS; j++)
		{
			mole_board[i][j].mole_present = FALSE;
			mole_board[i][j].been_hit = FALSE;
		}
}

/***************************************************
  
	MOLE MANAGEMENT ROUTINES 

 ***************************************************/


void add_mole(mole mole_board[NUM_ROWS][NUM_COLUMNS])
{
	int row, col;
	int bad;

	row = rand() % NUM_ROWS;
	col = rand() % NUM_COLUMNS;

	/* Add the mole */
	mole_board[row][col].mole_present = TRUE;
	mole_board[row][col].been_hit = FALSE;

	/* Give a 1 in 4 chance of getting a bad mole */
	bad = rand() % 4;
	if (bad == 1) mole_board[row][col].bad_mole = TRUE;
	else mole_board[row][col].bad_mole = FALSE;

	/* Give the mole a duration from 2 to 22 (.2 to 2.5 seconds) */
	mole_board[row][col].duration = (rand() % 20) + 2;
}


/* Decrements the timer for each mole on the board. If a moles timer
 * runs out, erase it from the board */
void handle_mole_timer(mole mole_board[NUM_ROWS][NUM_COLUMNS])
{
	int i, j;

	for (i = 0; i < NUM_ROWS; i++)
		for (j = 0; j < NUM_COLUMNS; j++)
		{
			if (mole_board[i][j].mole_present == TRUE)
			{
				mole_board[i][j].duration--;
				if (mole_board[i][j].duration <= 0)
				{
					mole_board[i][j].mole_present = FALSE;
					mole_board[i][j].been_hit = FALSE;
				}
			}
		}
} 

/*****************************************************************************

     BUTTONS AND MOUSE CLICKS

*****************************************************************************/


/* 
 * Draws the buttons.
 */
void draw_buttons()
{
	int i;
	int text_offset = MENU_LEFT+10;	/* offset of button text in button */
	int font_size = 12;				/* size of button text */
	
	/* draw rectangles for each button */
	for(i = 0; i < NUM_BUTTONS; i++) {
		GP142_rectangleXY(BLACK,
			MENU_LEFT,
			MENU_TOP - i * BUTTON_HEIGHT,
			MENU_LEFT + BUTTON_WIDTH,
			MENU_TOP - (i + 1)*BUTTON_HEIGHT, 1);
	}
	
	/* display button text */
	i = MENU_TOP - (BUTTON_HEIGHT/2) - (font_size/2);
	GP142_printfXY(FOREST_GREEN, text_offset, i, font_size, "New Game");	
	
	i -= BUTTON_HEIGHT;
	GP142_printfXY(RED, text_offset, i, font_size, "Exit");
	
}

/* 
 * Figures out where the mouse was clicked; on a button or on the canvas.
 */
int  classify_mouse_click(int mouse_x, int mouse_y)
{
	if ((mouse_x > MENU_LEFT) &&
		(mouse_x < MENU_LEFT + BUTTON_WIDTH) &&
		(mouse_y < MENU_TOP) &&
		(mouse_y > (MENU_TOP - NUM_BUTTONS*BUTTON_HEIGHT)))
		return BUTTON_CLICK;
	else
		return OTHER_CLICK;
}



/* 
 * Process mouse click on the buttons. Restarts the game or exits.
 */
void handle_button(int x, int y, mole mole_board[NUM_ROWS][NUM_COLUMNS], 
				   int *lives, int *score, int *quit)
{
	int buttonNum;		/* number of selected button */
	
	/* Convert y coordinate to button number from top. */
	buttonNum = (MENU_TOP - y)/BUTTON_HEIGHT;	  
	
	switch(buttonNum) 
	{
		/* New Game button */
		case 0:
			initialize_board(mole_board);
			*score = 0;
			*lives = 3;
			break;

		/* Exit button */
		case 1:
			*quit = TRUE;
			break;

		default:
			/*Bad news if this is ever reached! */
			something_is_wrong();
			break;
	}
}





/*
 * Process button click at location x, y.  Check for a mole
 * and respond accordingly. For a good mole, erase it and increment
 * our score. For a bad mole, erase it and decrement our lives. 
 */
void handle_canvas(int x, int y, mole moles[NUM_ROWS][NUM_COLUMNS],
				   int *lives, int *score)
{
	int row, col;
	
	if(   (x >= BOARD_MINX && x <= BOARD_MAXX)
		&& (y >= BOARD_MINY && y <= BOARD_MAXY))
	   {
		col = (x - BOARD_MINX) / CELL_HEIGHT;
		row = (y - BOARD_MINY) / CELL_WIDTH;
		
		if (moles[row][col].mole_present == TRUE && moles[row][col].been_hit == FALSE)
		{
			if (moles[row][col].bad_mole)
				*lives = *lives - 1;
			else *score = *score + 1000;
			
			/* Set the duration to 5 (.5 seconds) and been_hit to true
			* so that the mole will be drawn as HIT for half a second */
			moles[row][col].been_hit = TRUE;
			moles[row][col].duration = 5;
		}
	}
}


/*****************************************************************************

     OTHER

*****************************************************************************/

/* 
 * Display error message if internal error is detected. 
 */
void something_is_wrong (void) {
	GP142_printfXY (RED, 0, 0, 16, "BAAAADDD NEWS!");
}